import 'dart:collection';
import 'dart:math';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/scheduler.dart';
import 'package:flutter/widgets.dart';

import 'morph_transition.dart';

const int _kAnimationSpeed = 500;

const Duration _kRemovingDuration =
    const Duration(milliseconds: _kAnimationSpeed);
const Duration _kResizingDuration =
    const Duration(milliseconds: _kAnimationSpeed);
const Duration _kInsertingDuration =
    const Duration(milliseconds: _kAnimationSpeed);
const Duration _kChangingDuration =
    const Duration(milliseconds: _kAnimationSpeed);

const Curve _kRemovingCurve = Curves.easeOut;
const Curve _kResizingCurve = Curves.easeInOut;
const Curve _kInsertingCurve = Curves.easeIn;
const Curve _kChangingCurve = Curves.linear;

/// Provide durations and curves for all animations of the [AnimatedSliverList].
class AnimatedListAnimationSettings {
  final Duration removingDuration,
      resizingDuration,
      insertingDuration,
      changingDuration;
  final Curve removingCurve, resizingCurve, insertingCurve, changingCurve;

  const AnimatedListAnimationSettings({
    this.removingDuration = _kRemovingDuration,
    this.resizingDuration = _kResizingDuration,
    this.insertingDuration = _kInsertingDuration,
    this.changingDuration = _kChangingDuration,
    this.removingCurve = _kRemovingCurve,
    this.resizingCurve = _kResizingCurve,
    this.insertingCurve = _kInsertingCurve,
    this.changingCurve = _kChangingCurve,
  });
}

/// Use this controller to dispatch the changes of your list to the [AnimatedSliverList].
class AnimatedListController {
  AnimatedRenderSliverList _adaptor;

  AnimatedRenderSliverList get adaptor => _adaptor;

  void _checkAdaptor() {
    if (_adaptor == null)
      throw "This controller has not been assigned to an AnimatedListSliver widget yet.";
  }

  void _setAdaptor(AnimatedRenderSliverList adaptor) {
    if (_adaptor == adaptor) return;
    if (_adaptor != null) {
      throw "This controller cannot be assigned to multiple AnimatedListSliver widgets.";
    }
    _adaptor = adaptor;
  }

  /// Notify the [AnimatedSliverList] that a new range starting from [from] and [count] long
  /// has been inserted. Call this method after actually you have updated your list.
  void notifyInsertedRange(int from, int count) {
    _checkAdaptor();
    _adaptor._notifyReplacedRange(from, 0, count, null);
  }

  /// Notify the [AnimatedSliverList] that a range starting from [from] and [count] long
  /// has been removed. Call this method after actually you have updated your list.
  /// A new builder [removeItemBuilder] has to be provided in order to build the removed
  /// items when animating.
  void notifyRemovedRange(
      int from, int count, NullableIndexedWidgetBuilder removeItemBuilder) {
    _checkAdaptor();
    _adaptor._notifyReplacedRange(from, count, 0, removeItemBuilder);
  }

  /// Notify the [AnimatedSliverList] that a range starting from [from] and [count] long
  /// has been replaced with a new [newCount] long range. Call this method after
  /// you have updated your list.
  /// A new builder [removeItemBuilder] has to be provided in order to build the replaced
  /// items when animating.
  void notifyReplacedRange(int from, int removeCount, int insertCount,
      NullableIndexedWidgetBuilder removeItemBuilder) {
    _checkAdaptor();
    _adaptor._notifyReplacedRange(
        from, removeCount, insertCount, removeItemBuilder);
  }

  /// Notify the [AnimatedSliverList] that a range starting from [from] and [count] long
  /// has been modified. Call this method after actually you have updated your list.
  /// A new builder [changeItemBuilder] has to be provided in order to build the old
  /// items when animating.
  void notifyChangedRange(
      int from, int count, NullableIndexedWidgetBuilder changeItemBuilder) {
    _checkAdaptor();
    _adaptor._notifyChangedRange(from, count, changeItemBuilder);
  }

  /// After notifying your changes, call this method in order to dispatch them to the
  /// [AnimatedSliverList]. Eventually the animations that follow will start.
  void dispatchChanges() {
    _checkAdaptor();
    _adaptor._dispatchChanges();
  }
}

/// Abstract class for building animated Widgets.
abstract class AnimatedListAnimationBuilder {
  /// Wraps the [widget] in an animated widget for removing items.
  Widget buildRemoving(
      BuildContext context, Widget widget, Animation<double> animation);

  /// Wraps the [widget] in an animated widget for inserting new items.
  Widget buildInserting(
      BuildContext context, Widget widget, Animation<double> animation);

  /// Wraps the old Widget [fromWidget] and to new Widget [toWidget] in an animated one for changing items.
  Widget buildChanging(BuildContext context, Widget fromWidget, Widget toWidget,
      Animation<double> animation);
}

/// Default singleton class for building animated Widgets.
///
/// For removing items a fade out effect will be provided.
/// For inserting new items a fade in effect will be provided.
/// For changing items a morph effect will be provided.
class DefaultAnimatedListAnimationBuilder
    implements AnimatedListAnimationBuilder {
  const DefaultAnimatedListAnimationBuilder._();

  static const DefaultAnimatedListAnimationBuilder instance =
      const DefaultAnimatedListAnimationBuilder._();

  @override
  Widget buildRemoving(final BuildContext context, final Widget widget,
      final Animation<double> animation) {
    return FadeTransition(
        child: widget, opacity: Tween(begin: 1.0, end: 0.0).animate(animation));
  }

  @override
  Widget buildInserting(final BuildContext context, final Widget widget,
      final Animation<double> animation) {
    return FadeTransition(child: widget, opacity: animation);
  }

  @override
  Widget buildChanging(final BuildContext context, final Widget fromWidget,
      final Widget toWidget, final Animation<double> animation) {
    return MorphTransition(
        fromChild: fromWidget, toChild: toWidget, animation: animation);
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedSliverList (Widget)
//---------------------------------------------------------------------------------------------

/// An animated version of the standard [SliverList].
///
/// A new delegate version of [SliverChildBuilderDelegate] called [AnimatedSliverChildBuilderDelegate]
/// has to be provided.
///
/// The [AnimatedListController] controller should also be  provided to notify this sliver
/// when a change in the underlying list occurs.
///
/// A custom [AnimatedListAnimationBuilder] can be also provided to customize animations.
///
/// A [AnimatedListAnimationSettings] can be also provided to customize durations and curves of the animations.
///
/// The [updateScrollableWhenResizing] attribute if is set to [false] prevents scroll listeners for being
/// notified when a resize animation occurs.
///
/// The [coordinateAnimations] attribute if is set to [false] to prevents coordination of animations. If [true]
/// (as default) all animations will follow some priorities: intervals in removal state will start first, followed
/// by those in resizing and changing state, and finally those in inserting state.
class AnimatedSliverList extends SliverWithKeepAliveWidget {
  final AnimatedSliverChildBuilderDelegate delegate;
  final AnimatedListController controller;
  final AnimatedListAnimationBuilder animationBuilder;
  final AnimatedListAnimationSettings animatedListAnimationSettings;
  final bool updateScrollableWhenResizing;
  final bool coordinateAnimations;

  const AnimatedSliverList({
    Key key,
    this.delegate,
    this.controller,
    this.animationBuilder = DefaultAnimatedListAnimationBuilder.instance,
    this.animatedListAnimationSettings = const AnimatedListAnimationSettings(),
    this.updateScrollableWhenResizing = true,
    this.coordinateAnimations = true,
  }) : super(key: key);

  @override
  AnimatedSliverMultiBoxAdaptorElement createElement() =>
      AnimatedSliverMultiBoxAdaptorElement(this);

  @override
  AnimatedRenderSliverList createRenderObject(BuildContext context) {
    return AnimatedRenderSliverList(childManager: context, widget: this);
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedRenderSliverList (Render Object)
//---------------------------------------------------------------------------------------------

typedef _IntervalCreationCallback = _Interval Function(
    int from,
    int newRemoveCount,
    int newInsertCount,
    NullableIndexedWidgetBuilder removeItemBuilder,
    [_AnimatedListIntervalEventCallback onDisposed]);

typedef _AnimationItemBuilder = Widget Function(
    BuildContext context,
    NullableIndexedWidgetBuilder oldBuilder,
    int oldIndex,
    NullableIndexedWidgetBuilder newBuilder,
    int newIndex,
    Animation<double> animation);

/// Render class of the [AnimatedSliverList].
/// This is an extension of the original class [RenderSliverList].
class AnimatedRenderSliverList extends RenderSliverList
    implements TickerProvider {
  AnimatedSliverList get widget => childManager.widget;

  _IntervalList _intervals = _IntervalList();
  Set<Ticker> _tickers;

  /// Creates a sliver that places multiple box children in a linear array along
  /// the main axis, taking in account of animations resulting from changes to the underlying list.
  ///
  /// The [childManager] argument must not be null.
  AnimatedRenderSliverList({
    @required AnimatedSliverMultiBoxAdaptorElement childManager,
    AnimatedSliverList widget,
  }) : super(childManager: childManager);

  String toDebugString() => _intervals.toDebugString();

  /// Updates this render object against a new [AnimatedSliverList] Widget.
  void update(AnimatedSliverList widget) {
    widget.controller._setAdaptor(this);
    _intervals.clear();
    _tickers?.forEach((t) => _removeTicker(t));
    _tickers?.clear();
  }

  /// Creates a new ticker for animations.
  @override
  Ticker createTicker(final TickerCallback onTick) {
    var onTickNew = (Duration elapsed) {
      _markNeedsLayoutIfResizing();
      onTick(elapsed);
    };
    _tickers ??= <_WidgetTicker>{};
    final _WidgetTicker result = _WidgetTicker(onTickNew, this);
    _tickers.add(result);
    return result;
  }

  /// Marks this render object to relayout its children when animating resizing or changing intervals.
  void _markNeedsLayoutIfResizing() {
    for (final _Interval interval in _intervals) {
      if (interval.isInResizingState || interval.isInChangingState) {
        markNeedsLayout();
        return;
      }
    }
  }

  /// Removes a ticker.
  void _removeTicker(_WidgetTicker ticker) {
    assert(_tickers != null);
    assert(_tickers.contains(ticker));
    _tickers.remove(ticker);
  }

  /// Disposing method that clears all previously created tickers.
  void dispose() {
    _intervals.clear();
    assert(() {
      if (_tickers != null) {
        for (final Ticker ticker in _tickers) {
          if (ticker.isActive) {
            throw '$this was disposed with an active Ticker.';
          }
        }
      }
      return true;
    }());
  }

  /// A [didChangeDependencies] method like in [State.didChangeDependencies], necessary to update
  /// the tickers correctly when the "muted" attribute changes.
  void didChangeDependencies(final BuildContext context) {
    final bool muted = !TickerMode.of(context);
    if (_tickers != null) {
      for (final Ticker ticker in _tickers) {
        ticker.muted = muted;
      }
    }
  }

  /// This method is responsible for creating the correct Widget for a given child's index.
  Widget _build(final BuildContext context, final int index) {
    int adj = 0;
    for (int j = 0; j < _intervals.length; j++) {
      var interval = _intervals[j];

      switch (interval._state) {
        case _IntervalState.REMOVING:
          if (index >= interval._index) {
            if (index < (interval._index + interval._removeCount)) {
              return AbsorbPointer(
                  child: widget.animationBuilder.buildRemoving(
                context,
                interval.removeItemBuilder(context, index - interval._index),
                interval.animation,
              ));
            } else {
              adj += interval._insertCount - interval._removeCount;
            }
          }
          break;
        case _IntervalState.CHANGING:
          if (index >= interval._index) {
            if (index < (interval._index + interval._removeCount)) {
              return AbsorbPointer(
                  child: widget.animationBuilder.buildChanging(
                context,
                interval.removeItemBuilder(context, index - interval._index),
                _defaultInsertItemBuilder(
                    context, index - interval._index, interval),
                interval.animation,
              ));
            } else {
              adj += interval._insertCount - interval._removeCount;
            }
          }
          break;
        case _IntervalState.RESIZING:
          if (index == interval._index) {
            return Container(key: ValueKey(interval));
          } else if (index > interval._index) {
            adj += interval._insertCount - 1;
          }
          break;
        case _IntervalState.INSERTING:
          if (index >= interval._index &&
              index < (interval._index + interval._insertCount)) {
            return AbsorbPointer(
                child: widget.animationBuilder.buildInserting(
              context,
              _defaultInsertItemBuilder(
                  context, index - interval._index, interval),
              interval.animation,
            ));
          }
          break;
        default:
          throw "Illegal state";
      }
    }

    /// Create a still (not animating) widget
    var w = _animatedBuild(context, index + adj, false);

    // we insert a invisible boundary item at the end of the list in order to have
    // always an item in the viewport on the borders between this sliver and the next, so
    // to avoid a recalculation of the items in performLayout from scratch, thus causing
    // an annoying scroll jump due to an estimation error in calculating the size of
    // one or more resizing intervals
    if (w == null && _animatedBuild(context, index + adj - 1, true) != null) {
      return SizedBox.shrink();
    }

    return w;
  }

  /// Delegates the creation of the item Widget to the [AnimatedSliverChildBuilderDelegate.build] method.
  Widget _animatedBuild(BuildContext context, int index, bool animating) {
    childManager._animating = animating;
    return widget.delegate.build(context, index);
  }

  bool _notifying = false;

  /// Called from [AnimatedListController] when the user notified that a range of its list has been
  /// replaced by a new one. This method is also called in case of a simple insertion or removal.
  void _notifyReplacedRange(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder) {
    if (removeCount == 0 && insertCount == 0) return;

    _notifying = true;

    _insertAndAdjust(from, removeCount, insertCount, removeItemBuilder,
        _insertNewInterval, false);
  }

  /// Called from [AnimatedListController] when the user notified that a range of its list has been
  /// changed.
  void _notifyChangedRange(final int from, final int count,
      final NullableIndexedWidgetBuilder changeItemBuilder) {
    if (count == 0) return;

    _notifying = true;

    _insertAndAdjust(from, count, count, changeItemBuilder,
        _insertNewChangingInterval, true);
  }

  /// Called from [AnimatedListController] when the user has finished notifying any changes of its list.
  /// All pending notification will be optimized first, then all animations that follow will start.
  void _dispatchChanges() {
    _optimize();
    _notifying = false;
    _startAnimations();
  }

  /// Returns [true] if two intervals can ben merged for optimization.
  bool _canBeIntervalsMerged(_Interval a, _Interval b) =>
      a._index + a.buildingItemCount == b._index &&
      a._waiting &&
      b._waiting &&
      (a.isInRemovingState &&
              (b.isInRemovingState ||
                  b.isInResizingState && b._fromSize == 0.0) ||
          a.isInResizingState &&
              a._fromSize == 0.0 &&
              (b.isInRemovingState || b.isInResizingState) ||
          a.isInChangingState && b.isInChangingState);

  /// Merges two intervals for optimization.
  bool _mergeIntervals(_Interval a, _Interval b) {
    assert(_canBeIntervalsMerged(a, b));
    if (a._state == b._state) {
      final ra = a.removeItemBuilder;
      final rb = b.removeItemBuilder;
      final n = a._removeCount;
      a._removeCount += b._removeCount;
      a._insertCount += b._insertCount;
      a._toSize = (a._insertCount == 0) ? 0.0 : null;
      if (!a.isInResizingState)
        a.removeItemBuilder =
            (c, i) => (i < n) ? ra.call(c, i) : rb.call(c, i - n);
      return false;
    } else if (b.isInResizingState) {
      assert(a.isInRemovingState);
      a._insertCount += b._insertCount;
      a._toSize = (a._insertCount == 0) ? 0.0 : null;
      return false;
    } else {
      assert(a.isInResizingState && b.isInRemovingState);
      b._insertCount += a._insertCount;
      b._toSize = (b._insertCount == 0) ? 0.0 : null;
      b._index = a._index;
      return true;
    }
  }

  /// Optimize the created intervals, merging some or all of them if possibile.
  void _optimize() {
    _Interval s, i;
    for (int j = _intervals.length - 1; j >= 0; j--) {
      i = _intervals[j];
      if (s != null) {
        if (_canBeIntervalsMerged(i, s)) {
          final int n = i.buildingItemCount + s.buildingItemCount;
          int m;
          if (_mergeIntervals(i, s)) {
            m = s.buildingItemCount;
            i.dispose();
            i = s;
            _intervals.removeAt(j);
          } else {
            m = i.buildingItemCount;
            s.dispose();
            _intervals.removeAt(j + 1);
          }
          if (n != m) {
            for (int k = j + 1; k < _intervals.length; k++) {
              _intervals[k]._index += m - n;
            }
          }
        }
      }
      s = i;
    }
  }

  /// Core method for adapting or creating new intervals on a new user notification.
  void _insertAndAdjust(
    int from,
    int removeCount,
    int insertCount,
    final NullableIndexedWidgetBuilder removeItemBuilder,
    final _IntervalCreationCallback callback,
    final bool animating,
  ) {
    int adjust = 0;
    int rba = 0;
    for (int i = 0; i < _intervals.length; i++) {
      final int to = from + removeCount;

      final _Interval interval = _intervals[i];
      final int ifrom = interval._index + adjust;
      final int ito = ifrom + interval._insertCount;

      //  from       to                          ifrom       ito
      //   v          v                            v          v
      //   +++++++++++  (new rem/chg interval)     ooooooooooo  (existing final (after inserting) interval)
      // Note: X indicates overlap

      int adjustInterval(
          int rem, int ins, int leading, int trailing, int rbaAdjust) {
        return _adjustInterval(
            interval,
            rem,
            ins,
            leading,
            trailing,
            _shiftItemBuilder(removeItemBuilder, rba + rbaAdjust),
            callback,
            animating);
      }

      if (from < ifrom) {
        if (to <= ifrom) {
          // [+++  ooooo] -> [+++++ooooo]
          break;
        } else if (to <= ito) {
          // [+++++XXXoo] -> [+++++XXXXX]
          final int leading = ifrom - from;
          final int rem = removeCount - leading;
          final int ins = min(rem, insertCount);
          adjustInterval(rem, ins, 0, ito - to, leading);
          insertCount -= ins;
          removeCount -= rem;
          break;
        } else {
          // [+++++XXXXX+++++]
          final int overlapSize = ito - ifrom;
          final int leading = ifrom - from;
          int rem = leading;
          int ins = min(rem, insertCount);
          callback(from - adjust, rem, ins,
              _shiftItemBuilder(removeItemBuilder, rba));
          i++;
          rba += leading;
          insertCount -= ins;
          removeCount -= rem;
          rem = overlapSize;
          ins = min(rem, insertCount);
          final adj = interval.adjustingItemCount;
          final a = adjustInterval(rem, ins, 0, 0, 0);
          if (a >= 0) {
            adjust += adj;
            from = ito;
          }
          i += a;
          rba += overlapSize;
          insertCount -= ins;
          removeCount -= rem;
          continue;
        }
      } else if (from < ito) {
        if (to <= ito) {
          // [ooXXoo] -> [XXXXXX]
          adjustInterval(removeCount, insertCount, from - ifrom, ito - to, 0);
          return;
        } else {
          // [ooXXXX++++++] -> [XXXXXX++++++]
          final int rem = ito - from;
          final int ins = min(rem, insertCount);
          final adj = interval.adjustingItemCount;
          final a = adjustInterval(rem, ins, from - ifrom, 0, 0);
          if (a >= 0) {
            adjust += adj;
            from = ito;
          }
          i += a;
          rba += rem;
          insertCount -= ins;
          removeCount -= rem;
          continue;
        }
      } else {
        // [oooooo++++++++] -> [oooooo   ++++++++]
      }
      adjust += interval.adjustingItemCount;
    }

    callback(from - adjust, removeCount, insertCount,
        _shiftItemBuilder(removeItemBuilder, rba));
  }

  /// Called by [_insertAndAdjust] to adapt an existing interval in order to comply with the new notification.
  int _adjustInterval(
      final _Interval interval,
      final int newRemoveCount,
      final int newInsertCount,
      final int leading,
      final int trailing,
      final NullableIndexedWidgetBuilder removeItemBuilder,
      final _IntervalCreationCallback callback,
      final bool changing) {
    switch (interval._state) {
      case _IntervalState.REMOVING:
        interval._insertCount += newInsertCount - newRemoveCount;
        interval._toSize = (interval._insertCount == 0) ? 0.0 : null;
        break;
      case _IntervalState.RESIZING:
        interval._insertCount += newInsertCount - newRemoveCount;
        return interval.resize() ? -1 : 0;
      case _IntervalState.INSERTING:
        return _splitInterval(
          interval,
          leading,
          trailing,
          newRemoveCount,
          newInsertCount,
          removeItemBuilder,
          callback,
          (context, oldBuilder, oldIndex, newBuilder, newIndex, animation) {
            return widget.animationBuilder.buildInserting(
                context, newBuilder.call(context, newIndex), animation);
          },
        );
        break;
      case _IntervalState.CHANGING:
        if (changing && interval._waiting) return 0;
        return _splitInterval(
          interval,
          leading,
          trailing,
          newRemoveCount,
          newInsertCount,
          removeItemBuilder,
          callback,
          (context, oldBuilder, oldIndex, newBuilder, newIndex, animation) {
            return widget.animationBuilder.buildChanging(
                context,
                oldBuilder.call(context, oldIndex),
                newBuilder.call(context, newIndex),
                animation);
          },
        );
        break;
      default:
        break;
    }
    return 0;
  }

  /// Called by [_adjustInterval] to eventually split an existing interval in more parts.
  int _splitInterval(
    final _Interval interval,
    final int leading,
    final int trailing,
    final int newRemoveCount,
    final int newInsertCount,
    final NullableIndexedWidgetBuilder itemBuilder,
    final _IntervalCreationCallback newIntervalBuilder,
    final _AnimationItemBuilder animationItemBuilder,
  ) {
    assert(interval.isInChangingState || interval.isInInsertingState);

    if (newRemoveCount == 0 && newInsertCount == 0) return 0;

    final int from = interval._index + leading;

    final w = interval._waiting;

    final oldBuilder = interval.removeItemBuilder;

    final _AnimationController orphan =
        (!interval._waiting && newRemoveCount > 0)
            ? _AnimationController.clone(this, interval._controller)
            : null;

    var other = interval.split(this, leading, trailing);
    if (other != null) _intervals.insert(other);

    var q = newIntervalBuilder.call(
      from,
      newRemoveCount,
      newInsertCount,
      newRemoveCount == 0
          ? null
          : (context, index) => animationItemBuilder.call(
              context,
              oldBuilder,
              index + leading,
              itemBuilder,
              index,
              orphan?._animation ?? kAlwaysDismissedAnimation),
      (disposedInterval, beforeUpdatingState) {
        if (!beforeUpdatingState) orphan?.dispose();
      },
    );
    if (!w) q.startAnimation();

    return leading == 0 ? 0 : 1;
  }

  /// Called by [_insertAndAdjust] to create a new interval in initial removing or resizing status.
  _Interval _insertNewInterval(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder,
      [final _AnimatedListIntervalEventCallback onDisposed]) {
    if (removeCount == 0 && insertCount == 0) return null;
    _Interval interval;
    interval = _Interval(
      vsync: this,
      animationSettings: widget.animatedListAnimationSettings,
      index: from,
      insertCount: insertCount,
      removeCount: removeCount,
      removeItemBuilder: removeItemBuilder,
      onRemovingCompleted: (interval, b) =>
          _onIntervalRemovingCompleted(interval, b),
      onResizingCompleted: (interval, b) =>
          _onIntervalResizingCompleted(interval, b),
      onInsertingCompleted: (interval, b) =>
          _onIntervalInsertingCompleted(interval, b),
      onCompleted: (interval, b) => _onIntervalCompleted(interval, b),
      onDisposed: onDisposed,
    );
    int index = _intervals.insert(interval);
    if (interval._removeCount == 0)
      while (++index < _intervals.length) _intervals[index]._index++;
    childManager._onUpdateOnNewInterval(interval);
    return interval;
  }

  /// Called by [_insertAndAdjust] to create a new interval in initial changing status.
  _Interval _insertNewChangingInterval(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder,
      [final _AnimatedListIntervalEventCallback onDisposed]) {
    assert(removeCount == insertCount);
    if (removeCount == 0 && insertCount == 0) return null;
    _Interval interval;
    interval = _Interval.change(
      vsync: this,
      animationSettings: widget.animatedListAnimationSettings,
      index: from,
      changeCount: removeCount,
      removeItemBuilder: removeItemBuilder,
      onChangingCompleted: (interval, b) =>
          _onIntervalChangingCompleted(interval, b),
      onCompleted: (interval, b) => _onIntervalCompleted(interval, b),
      onDisposed: onDisposed,
    );
    _intervals.insert(interval);
    childManager._onUpdateOnNewInterval(interval);
    return interval;
  }

  /// Called when interval states change to begin new animations.
  void _startAnimations() {
    if (!_notifying) _intervals.startAnimations(widget.coordinateAnimations);
  }

  /// Called when a removing interval is completed.
  void _onIntervalRemovingCompleted(
      final _Interval interval, final bool beforeUpdatingState) {
    if (beforeUpdatingState) {
      int i = _intervals.indexOf(interval);
      assert(i >= 0);
      while (++i < _intervals.length)
        _intervals[i]._index += 1 - interval._removeCount;
    } else {
      childManager._onUpdateOnIntervalRemovedToResizing(interval);
      _startAnimations();
    }
  }

  /// Called when a resizing interval is completed.
  void _onIntervalResizingCompleted(
      final _Interval interval, final bool beforeUpdatingState) {
    if (beforeUpdatingState) {
      int i = _intervals.indexOf(interval);
      assert(i >= 0);
      while (++i < _intervals.length)
        _intervals[i]._index += interval._insertCount - 1;
    } else {
      if (interval._insertCount == 0)
        childManager._onUpdateOnIntervalResizedToDisposing(interval);
      else
        childManager._onUpdateOnIntervalResizedToInserting(interval);
      _startAnimations();
    }
  }

  /// Called when a inserting interval is completed.
  void _onIntervalInsertingCompleted(
      final _Interval interval, final bool beforeUpdatingState) {
    if (!beforeUpdatingState) {
      childManager._onUpdateOnIntervalInsertedToDisposing(interval);
      _startAnimations();
    }
  }

  /// Called when a changing interval is completed.
  void _onIntervalChangingCompleted(
      final _Interval interval, final bool beforeUpdatingState) {
    if (!beforeUpdatingState) {
      childManager._onUpdateOnIntervalChangedToDisposing(interval);
      _startAnimations();
    }
  }

  /// Called when an interval has completed all animations.
  void _onIntervalCompleted(
      final _Interval interval, final bool beforeUpdatingState) {
    if (beforeUpdatingState) {
      final int index = _intervals.indexOf(interval);
      assert(index >= 0);
      _intervals.removeAt(index);
    } else {
      _startAnimations();
    }
  }

  // the inserting builder picks the item directly from the final changed list
  Widget _defaultInsertItemBuilder(
      final BuildContext context, int index, final _Interval interval) {
    index += interval._index;
    int i = _intervals.indexOf(interval);
    assert(i >= 0);
    for (int j = 0; j < i; j++) {
      var prevInterval = _intervals[j];
      index += prevInterval.adjustingItemCount;
    }
    return _animatedBuild(context, index, true);
  }

  /// Copied from [RenderSliverMultiBoxAdaptor.indexOf].
  /// An assertion was removed to accept the dummy child element.
  @override
  int indexOf(final RenderBox child) {
    assert(child != null);
    final SliverMultiBoxAdaptorParentData childParentData =
        child.parentData as SliverMultiBoxAdaptorParentData;
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    // assert(childParentData.index != null);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    return childParentData.index;
  }

  @override
  AnimatedSliverMultiBoxAdaptorElement get childManager => super.childManager;

  bool _inPerformLayout = false;

  /// This method has been overridden to eventually measure the size of new resizing intervals and
  /// to eventually adjust the scroll offset.
  /// If required by [AnimatedSliverList.updateScrollableWhenResizing] attribute, calls the
  /// [_notifyScrollable] method to update the scroll listeners when a resizing inerval is animating.
  @override
  void performLayout() {
    _inPerformLayout = true;

    bool resizing = false;
    final BoxConstraints childConstraints = constraints.asBoxConstraints();
    for (_Interval i in _intervals) {
      if (i.isInResizingState) {
        resizing = true;
        if (i._fromSize == null || i._toSize == null)
          _measureIntervalSize(i, childConstraints);
      }
    }

    super.performLayout();

    _inPerformLayout = false;

    childManager._adjustScrollOffset();

    if (resizing && widget.updateScrollableWhenResizing) _notifyScrollable();
  }

  /// Dispatch a "fake" change to the [ScrollPosition] to force the listeners (like a [ScrollBar])
  /// to refresh its state.
  void _notifyScrollable() {
    ScrollPosition position =
        Scrollable.of(childManager)?.widget?.controller?.position;
    if (position == null) return;
    ScrollUpdateNotification(
            metrics: position.copyWith(),
            context: childManager,
            scrollDelta: 0,
            dragDetails: null)
        .dispatch(childManager);
  }

  /// This method has been overridden in order to return the correct size of the resizing intervals.
  @protected
  double paintExtentOf(final RenderBox child) {
    var parentData = child.parentData as SliverMultiBoxAdaptorParentData;
    if (parentData.index != null && _inPerformLayout) {
      var key = childManager._childElements[indexOf(child)]?.widget?.key;
      if (key is ValueKey && key.value is _Interval) {
        var interval = key.value;
        if (interval != null && interval.isInResizingState)
          return _measureIntervalSize(
              interval, this.constraints.asBoxConstraints());
      }
    }
    return super.paintExtentOf(child);
  }

  /// Calculate the space required by the specified resizing interval.
  /// If the range has been measured already, the previously calculated value is returned.
  double _measureIntervalSize(
      final _Interval interval, final BoxConstraints childConstraints) {
    assert(interval != null && interval.isInResizingState);
    double calculate(final int n, final NullableIndexedWidgetBuilder builder,
        final String debugLabel) {
      double size = 0.0;
      int i;
      final double maxSize = constraints.viewportMainAxisExtent +
          (parent as RenderViewport).cacheExtent * 2;
      invokeLayoutCallback<SliverConstraints>(
          (final SliverConstraints constraints) {
        for (i = 0; i < n; i++) {
          if (size > maxSize) break;
          final Widget w = builder.call(childManager, i);
          final r = childManager._layoutDummyChild(w, childConstraints);
          size += paintExtentOf(r);
        }
        if (i < n) size *= (n / i);
        childManager._dropDummyChild();
      });
      return size;
    }

    interval._fromSize ??= calculate(
        interval._removeCount, interval.removeItemBuilder, "fromSize");
    interval._toSize ??= calculate(interval._insertCount,
        (c, i) => _defaultInsertItemBuilder(c, i, interval), "toSize");
    assert(interval._fromSize != null && interval._toSize != null);
    return interval._currentSize;
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedSliverMultiBoxAdaptorElement (Render Object Element / Child Manager)
//---------------------------------------------------------------------------------------------

/// An element that lazily builds children for a [SliverMultiBoxAdaptorWidget].
///
/// Implements [RenderSliverBoxChildManager], which lets this element manage
/// the children of subclasses of [RenderSliverMultiBoxAdaptor].
///
/// This class takes the implementation of the standard [SliverMultiBoxAdaptorElement] class
/// and provides support for animations in face of list changes.
class AnimatedSliverMultiBoxAdaptorElement extends RenderObjectElement
    implements RenderSliverBoxChildManager {
  /// Creates an element that lazily builds children for the given widget with support for
  /// animations in response to changes to the list.
  AnimatedSliverMultiBoxAdaptorElement(AnimatedSliverList widget)
      : super(widget);

  /// support for the method didChangeDependencies like in [State.didChangeDependencies]
  bool _didChangeDependencies = false;

  /// flag indicating that a reorganization of children is in progress beacuse an interval status has changed
  bool _remanagingChildren = false;

  /// flag passed to [AnimatedSliverChildBuilderDelegate.build] method.
  bool _animating;

  /// dummy element used to calculate the size of a resizing interval
  Element _dummyChild;

  /// list of updates to apply to this render object element in response to one or more changes to the list
  List<_BuildUpdate> _updates = List();

  /// This method has been overridden in order to handle the new [AnimatedRenderSliverList.didChangeDependecies]
  /// method like in [State.didChangeDependencies]. In addition the method attaches this instance
  /// to the [AnimatedListController].
  @override
  void mount(final Element parent, final dynamic newSlot) {
    super.mount(parent, newSlot);
    renderObject.didChangeDependencies(this);
    widget.controller?._setAdaptor(renderObject);
  }

  /// This method has been overridden in order to detach this instance from the [AnimatedListController] and
  /// to dispose the render object [AnimatedRenderSliverList].
  @override
  void unmount() {
    widget.controller?._setAdaptor(null);
    renderObject.dispose();
    super.unmount();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    _didChangeDependencies = true;
  }

  @override
  AnimatedSliverList get widget => super.widget as AnimatedSliverList;

  @override
  AnimatedRenderSliverList get renderObject =>
      super.renderObject as AnimatedRenderSliverList;

  /// Copied and adapted from [SliverMultiBoxAdaptorElement.update].
  @override
  void update(covariant AnimatedSliverList newWidget) {
    final AnimatedSliverList oldWidget = widget;
    super.update(newWidget);
    final AnimatedSliverChildBuilderDelegate newDelegate = newWidget.delegate;
    final AnimatedSliverChildBuilderDelegate oldDelegate = oldWidget.delegate;
    if (newDelegate != oldDelegate &&
        (newDelegate.runtimeType != oldDelegate.runtimeType ||
            newDelegate.shouldRebuild(oldDelegate))) {
      renderObject.update(widget); // update also the render object!
      performRebuild();
    }
  }

  SplayTreeMap<int, Element> _childElements = SplayTreeMap<int, Element>();
  RenderBox _currentBeforeChild;

  Widget _build(final int index) => renderObject._build(this, index);

  /// Copied from [SliverMultiBoxAdaptorElement.createChild].
  @override
  void createChild(final int index, {@required final RenderBox after}) {
    assert(_currentlyUpdatingChildIndex == null);
    owner.buildScope(this, () {
      final bool insertFirst = after == null;
      assert(insertFirst || _childElements[index - 1] != null);
      _currentBeforeChild = insertFirst
          ? null
          : (_childElements[index - 1].renderObject as RenderBox);
      Element newChild;
      try {
        _currentlyUpdatingChildIndex = index;
        newChild = updateChild(_childElements[index], _build(index), index);
      } finally {
        _currentlyUpdatingChildIndex = null;
      }
      if (newChild != null) {
        _childElements[index] = newChild;
      } else {
        _childElements.remove(index);
      }
    });
  }

  /// Copied from [SliverMultiBoxAdaptorElement.updateChild].
  @override
  Element updateChild(
      final Element child, final Widget newWidget, final dynamic newSlot) {
    final SliverMultiBoxAdaptorParentData oldParentData =
        child?.renderObject?.parentData as SliverMultiBoxAdaptorParentData;
    final Element newChild = super.updateChild(child, newWidget, newSlot);
    final SliverMultiBoxAdaptorParentData newParentData =
        newChild?.renderObject?.parentData as SliverMultiBoxAdaptorParentData;

    // Preserve the old layoutOffset if the renderObject was swapped out.
    if (oldParentData != newParentData &&
        oldParentData != null &&
        newParentData != null) {
      newParentData.layoutOffset = oldParentData.layoutOffset;
    }
    return newChild;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.forgetChild].
  @override
  void forgetChild(final Element child) {
    assert(child != null);
    assert(child.slot != null);
    assert(_childElements.containsKey(child.slot));
    _childElements.remove(child.slot);
    super.forgetChild(child);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.removeChild].
  @override
  void removeChild(final RenderBox child) {
    final int index = renderObject.indexOf(child);
    assert(_currentlyUpdatingChildIndex == null);
    assert(index >= 0);
    owner.buildScope(this, () {
      assert(_childElements.containsKey(index));
      try {
        _currentlyUpdatingChildIndex = index;
        final Element result = updateChild(_childElements[index], null, index);
        assert(result == null);
      } finally {
        _currentlyUpdatingChildIndex = null;
      }
      _childElements.remove(index);
      assert(!_childElements.containsKey(index));
    });
  }

  /// Estimates the max scroll offset based on the rendered viewport data.
  double _extrapolateMaxScrollOffset(
    final int firstIndex,
    final int lastIndex,
    final double leadingScrollOffset,
    final double trailingScrollOffset,
    final int childCount,
  ) {
    if (lastIndex == childCount - 1) return trailingScrollOffset;

    double innerSpace = 0.0, trailingSpace = 0.0;
    int innerCount = 0, trailingCount = 0;
    for (final _Interval interval in renderObject._intervals) {
      if (interval.isInResizingState) {
        if (firstIndex <= interval._index && interval._index <= lastIndex) {
          innerCount++;
          innerSpace += interval._currentSize;
        } else if (interval._index >= firstIndex) {
          trailingCount++;
          trailingSpace += interval._currentSize;
        }
      }
    }

    double ret = trailingScrollOffset + trailingSpace;

    final int remainingCount = childCount - lastIndex - 1 - trailingCount;
    if (remainingCount > 0) {
      final int reifiedCount = lastIndex - firstIndex + 1 - innerCount;
      double averageExtent;
      if (reifiedCount > 0) {
        averageExtent =
            (trailingScrollOffset - leadingScrollOffset - innerSpace) /
                reifiedCount;
      } else {
        double size = 0.0;
        int count = 0;
        for (final _Interval interval in renderObject._intervals) {
          if (interval.isInResizingState) {
            if (firstIndex <= interval._index && interval._index <= lastIndex) {
              assert(interval._fromSize != null && interval._toSize != null);
              size += interval._fromSize + interval._toSize;
              count += interval._removeCount + interval._insertCount;
            }
          }
        }
        averageExtent = size / count;
      }
      ret += averageExtent * remainingCount;
    }
    return ret;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.estimateMaxScrollOffset].
  /// The [AnimatedSliverChildBuilderDelegate.estimateMaxScrollOffset] method call has been removed just now.
  @override
  double estimateMaxScrollOffset(
    final SliverConstraints constraints, {
    final int firstIndex,
    final int lastIndex,
    final double leadingScrollOffset,
    final double trailingScrollOffset,
  }) {
    final int childCount = estimatedChildCount;
    if (childCount == null) return double.infinity;
    return
        /*widget.estimateMaxScrollOffset(
              constraints,
              firstIndex,
              lastIndex,
              leadingScrollOffset,
              trailingScrollOffset,
            ) ??*/
        _extrapolateMaxScrollOffset(
      firstIndex,
      lastIndex,
      leadingScrollOffset,
      trailingScrollOffset,
      childCount,
    );
  }

  int get estimatedChildCount => widget.delegate.estimatedChildCount;

  /// Copied from [SliverMultiBoxAdaptorElement.childCount].
  /// The final count will be changed to take in account the last invisibile boundary item and
  /// the items taken by all intervals.
  @override
  int get childCount {
    int result = estimatedChildCount;
    if (result == null) {
      // Since childCount was called, we know that we reached the end of
      // the list (as in, _build return null once), so we know that the
      // list is finite.
      // Let's do an open-ended binary search to find the end of the list
      // manually.
      int lo = 0;
      int hi = 1;
      const int max = kIsWeb
          ? 9007199254740992 // max safe integer on JS (from 0 to this number x != x+1)
          : ((1 << 63) - 1);
      while (renderObject._animatedBuild(this, hi - 1, true) != null) {
        lo = hi - 1;
        if (hi < max ~/ 2) {
          hi *= 2;
        } else if (hi < max) {
          hi = max;
        } else {
          throw FlutterError(
              'Could not find the number of children in ${widget.delegate}.\n'
              'The childCount getter was called (implying that the delegate\'s builder returned null '
              'for a positive index), but even building the child with index $hi (the maximum '
              'possible integer) did not return null. Consider implementing childCount to avoid '
              'the cost of searching for the final child.');
        }
      }
      while (hi - lo > 1) {
        final int mid = (hi - lo) ~/ 2 + lo;
        if (renderObject._animatedBuild(this, mid - 1, true) == null) {
          hi = mid;
        } else {
          lo = mid;
        }
      }
      result = lo;
    }
    assert(result != null);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    return result + 1 + renderObject._intervals.itemCountAdjustment;
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
  }

  SliverMultiBoxAdaptorParentData _parentDataOf(final int index) =>
      (_childElements[index].renderObject.parentData
          as SliverMultiBoxAdaptorParentData);

  /// Updates the [Geometry] is necessary to correct the scroll offset taking into account
  /// of the space changed when resizing intervals from the last frame.
  /// The rendered items will also be scrolled up or down according to it.
  void _adjustScrollOffset() {
    assert(debugAssertChildListLocked());

    double _scrollOffsetCorrection = 0.0;

    final int firstIndex = _childElements.firstKey() ?? 0;
    final int lastIndex = _childElements.lastKey() ?? 0;

    if (_childElements.isNotEmpty) {
      for (final _Interval interval in renderObject._intervals) {
        if (interval._index < firstIndex &&
            interval.isInResizingState &&
            interval._lastSize != null) {
          _scrollOffsetCorrection += interval._currentSize - interval._lastSize;
        }
        interval._lastSize = interval._currentSize;
      }

      final double firstOffset = _parentDataOf(firstIndex).layoutOffset;
      if (firstOffset + _scrollOffsetCorrection < 0.0) {
        _scrollOffsetCorrection = -firstOffset;
      }

      if (_scrollOffsetCorrection != 0.0) {
        for (int index = firstIndex; index <= lastIndex; index++) {
          _parentDataOf(index).layoutOffset += _scrollOffsetCorrection;
        }
      }
    }

    _scrollOffsetCorrection +=
        (renderObject.geometry.scrollOffsetCorrection ?? 0);
    if (_scrollOffsetCorrection != 0.0) {
      renderObject.geometry = SliverGeometry(
        scrollExtent: renderObject.geometry.scrollExtent,
        paintExtent: renderObject.geometry.paintExtent,
        paintOrigin: renderObject.geometry.paintOrigin,
        layoutExtent: renderObject.geometry.layoutExtent,
        maxPaintExtent: renderObject.geometry.maxPaintExtent,
        maxScrollObstructionExtent:
            renderObject.geometry.maxScrollObstructionExtent,
        hitTestExtent: renderObject.geometry.hitTestExtent,
        visible: renderObject.geometry.visible,
        hasVisualOverflow: renderObject.geometry.hasVisualOverflow,
        scrollOffsetCorrection: _scrollOffsetCorrection,
        cacheExtent: renderObject.geometry.cacheExtent,
      );
    }

    ////////////////////////////////////////////////////////
    // double so = (renderObject.constraints.scrollOffset ?? 0) +
    //     (renderObject.constraints.cacheOrigin ?? 0) +
    //     (renderObject.geometry.scrollOffsetCorrection ?? 0);
    // double sm = renderObject.geometry.scrollExtent ?? 0;
    // debugText =
    //     (so.isNaN || so.isInfinite || sm.isNaN || sm.isInfinite || sm == 0.0)
    //         ? "scroll UNDETERMINED"
    //         : "scroll ${so.toInt()}/${sm.toInt()}";
    ////////////////////////////////////////////////////////
  }

  //

  /// This method is being called from [AnimatedRenderSliverList] when creating a new interval.
  /// A new [_BuildUpdate] will be added to the update list and this element will be marked to rebuild.
  void _onUpdateOnNewInterval(final _Interval interval) {
    if (interval.isInRemovingState) {
      _updates.add(_BuildUpdate(
        index: interval._index,
        removeCount: interval._removeCount,
        insertCount: interval._insertCount,
        type: _BuildUpdateType.NEW_REMOVING_INTERVAL,
      ));
    } else if (interval.isInResizingState) {
      _updates.add(_BuildUpdate(
        index: interval._index,
        removeCount: interval._removeCount,
        insertCount: interval._insertCount,
        type: _BuildUpdateType.NEW_RESIZING_INTERVAL,
      ));
    }
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a removing interval becomes
  /// a resizing interval.
  /// A new [_BuildUpdate] will be added to the update list and this element will be marked to rebuild.
  void _onUpdateOnIntervalRemovedToResizing(final _Interval interval) {
    _updates.add(_BuildUpdate(
      index: interval._index,
      removeCount: interval._removeCount,
      insertCount: interval._insertCount,
      type: _BuildUpdateType.REMOVED_TO_RESIZING,
    ));
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a resizing interval becomes
  /// an inserting interval.
  /// A new [_BuildUpdate] will be added to the update list and this element will be marked to rebuild.
  void _onUpdateOnIntervalResizedToInserting(final _Interval interval) {
    _updates.add(_BuildUpdate(
      index: interval._index,
      removeCount: interval._removeCount,
      insertCount: interval._insertCount,
      type: _BuildUpdateType.RESIZED_TO_INSERTING,
    ));
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a resizing interval (without
  /// insertion) completes.
  /// A new [_BuildUpdate] will be added to the update list and this element will be marked to rebuild.
  void _onUpdateOnIntervalResizedToDisposing(final _Interval interval) {
    _updates.add(_BuildUpdate(
      index: interval._index,
      removeCount: interval._removeCount,
      insertCount: interval._insertCount,
      type: _BuildUpdateType.RESIZED_TO_DISPOSING,
    ));
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a inserting interval completes.
  /// This element will be marked to rebuild.
  void _onUpdateOnIntervalInsertedToDisposing(final _Interval interval) {
    // _updates.add(_BuildUpdate(
    //   index: interval.index,
    //   removeCount: interval.removeCount,
    //   insertCount: interval.insertCount,
    //   type: _BuildUpdateType.INSERTED_TO_DISPOSING,
    // ));
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a changing interval completes.
  /// This element will be marked to rebuild.
  void _onUpdateOnIntervalChangedToDisposing(final _Interval interval) {
    // _updates.add(_BuildUpdate(
    //   index: interval.index,
    //   removeCount: interval.removeCount,
    //   insertCount: interval.insertCount,
    //   type: _BuildUpdateType.CHANGED_TO_DISPOSING,
    // ));
    markNeedsBuild();
  }

  //

  /// Update this element according to a new created interval in removing status.
  /// Actually no operation is needed.
  void _updateOnNewRemovingInterval(final _BuildUpdate upd) {}

  /// Update this element according to a new created interval in resizing status.
  /// The last rendered children will be reorganizated in order to take in account of it.
  void _updateOnNewResizingInterval(final _BuildUpdate upd) {
    _remanagingChildren = true;
    assert(upd.removeCount == 0);
    final int firstIndex = _childElements.firstKey() ?? 0;
    final int lastIndex = _childElements.lastKey() ?? 0;
    for (int index = lastIndex; index >= firstIndex; index--) {
      if (index >= upd.index) _changeChildIndex(index, index + 1);
    }
    _remanagingChildren = false;
  }

  /// Updates this element according to a change of an interval from its removing status
  /// to resizing status.
  /// The last rendered children will be reorganizated in order to take in account of it.
  void _updateOnIntervalRemovedToResizing(final _BuildUpdate upd) {
    _remanagingChildren = true;
    final int firstIndex = _childElements.firstKey() ?? 0;
    final int lastIndex = _childElements.lastKey() ?? 0;
    for (int index = firstIndex; index <= lastIndex; index++) {
      if (index >= upd.index + upd.removeCount)
        _changeChildIndex(index, index - upd.removeCount + 1);
      else if (index >= upd.index) _removeChild(index);
    }
    _remanagingChildren = false;
  }

  /// Updates this element according to a change of an interval from its resizing status
  /// to inserting status.
  /// The last rendered children will be reorganizated in order to take in account of it.
  void _updateOnIntervalResizedToInserting(final _BuildUpdate upd) {
    _remanagingChildren = true;
    final int firstIndex = _childElements.firstKey() ?? 0;
    final int lastIndex = _childElements.lastKey() ?? 0;
    for (int index = lastIndex; index >= firstIndex; index--) {
      if (index > upd.index)
        _changeChildIndex(index, index + upd.insertCount - 1);
      if (index == upd.index) _removeChild(index);
    }
    _remanagingChildren = false;
  }

  /// Updates this element according to the completion of a resizing interval.
  /// The last rendered children will be reorganizated in order to take in account of it.
  void _updateOnIntervalResizedToDisposing(final _BuildUpdate upd) {
    assert(upd.insertCount == 0);
    _remanagingChildren = true;
    final int firstIndex = _childElements.firstKey() ?? 0;
    final int lastIndex = _childElements.lastKey() ?? 0;
    for (int index = firstIndex; index <= lastIndex; index++) {
      if (index > upd.index) _changeChildIndex(index, index - 1);
      if (index == upd.index) _removeChild(index);
    }
    _remanagingChildren = false;
  }

  /// Updates this element according to the completion of an inserting interval.
  // void _updateOnIntervalInsertedToDisposing(_BuildUpdate upd) {}

  /// Forse a rendered child to pass from an [oldIndex] to a [newIndex] position.
  void _changeChildIndex(final int oldIndex, final int newIndex) {
    if (oldIndex == newIndex) return;
    final Element child = _childElements[oldIndex];
    if (child == null) return;
    _currentlyUpdatingChildIndex = newIndex;
    updateSlotForChild(child, newIndex);
    (child.renderObject.parentData as SliverMultiBoxAdaptorParentData).index =
        newIndex;
    _childElements[newIndex] = child;
    _childElements.remove(oldIndex);
  }

  /// Removes the rendered child in the [index] position.
  void _removeChild(final int index) {
    final Element child = _childElements[index];
    if (child == null) return;
    (child.renderObject.parentData as SliverMultiBoxAdaptorParentData)
        .layoutOffset = null;
    _currentlyUpdatingChildIndex = index;
    _childElements[index] = updateChild(_childElements[index], null, index);
    _childElements.remove(index);
  }

  /// Copied and adapted from [SliverMultiBoxAdaptorElement.performRebuild].
  /// In addition:
  ///
  /// Calls the [AnimatedRenderSliverList.didChangeDependencies] method if
  /// a dependency has been changed.
  ///
  /// Apply all build updates [_BuildUpdate] of the list before calling the original
  /// method implementation.
  @override
  void performRebuild() {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if (_didChangeDependencies) {
      renderObject.didChangeDependencies(this);
      _didChangeDependencies = false;
    }
    super.performRebuild();
    for (final _BuildUpdate upd in _updates) {
      switch (upd.type) {
        case _BuildUpdateType.NEW_REMOVING_INTERVAL:
          _updateOnNewRemovingInterval(upd);
          break;
        case _BuildUpdateType.NEW_RESIZING_INTERVAL:
          _updateOnNewResizingInterval(upd);
          break;
        case _BuildUpdateType.REMOVED_TO_RESIZING:
          _updateOnIntervalRemovedToResizing(upd);
          break;
        case _BuildUpdateType.RESIZED_TO_INSERTING:
          _updateOnIntervalResizedToInserting(upd);
          break;
        case _BuildUpdateType.RESIZED_TO_DISPOSING:
          _updateOnIntervalResizedToDisposing(upd);
          break;
        // case _BuildUpdateType.INSERTED_TO_DISPOSING:
        //   _updateOnIntervalInsertedToDisposing(upd);
        //   break;
      }
    }
    _updates.clear();
    _currentlyUpdatingChildIndex = null;
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    _currentBeforeChild = null;
    assert(_currentlyUpdatingChildIndex == null);
    try {
      final SplayTreeMap<int, Element> newChildren =
          SplayTreeMap<int, Element>();
      final Map<int, double> indexToLayoutOffset = HashMap<int, double>();

      for (final int index in _childElements.keys.toList()) {
        assert(_childElements[index] != null);
        final Key key = _childElements[index].widget.key;
        int newIndex = key == null ? null : widget.delegate.findIndexByKey(key);

        final SliverMultiBoxAdaptorParentData childParentData =
            _childElements[index].renderObject?.parentData
                as SliverMultiBoxAdaptorParentData;

        if (childParentData != null && childParentData.layoutOffset != null)
          indexToLayoutOffset[index] = childParentData.layoutOffset;

        if (newIndex != null && newIndex != index) {
          // The layout offset of the child being moved is no longer accurate.
          if (childParentData != null) childParentData.layoutOffset = null;

          newChildren[newIndex] = _childElements[index];
          // We need to make sure the original index gets processed.
          newChildren.putIfAbsent(index, () => null);
          // We do not want the remapped child to get deactivated during processElement.
          _childElements.remove(index);
        } else {
          newChildren.putIfAbsent(index, () => _childElements[index]);
        }
      }

      void processElement(int index) {
        _currentlyUpdatingChildIndex = index;
        if (_childElements[index] != null &&
            _childElements[index] != newChildren[index]) {
          // This index has an old child that isn't used anywhere and should be deactivated.
          _childElements[index] =
              updateChild(_childElements[index], null, index);
        }
        final Element newChild =
            updateChild(newChildren[index], _build(index), index);
        if (newChild != null) {
          _childElements[index] = newChild;
          final SliverMultiBoxAdaptorParentData parentData = newChild
              .renderObject.parentData as SliverMultiBoxAdaptorParentData;
          if (index == 0) {
            parentData.layoutOffset = 0.0;
          } else if (indexToLayoutOffset.containsKey(index)) {
            parentData.layoutOffset = indexToLayoutOffset[index];
          }
          if (!parentData.keptAlive)
            _currentBeforeChild = newChild.renderObject as RenderBox;
        } else {
          _childElements.remove(index);
        }
      } // processElement

      // Moving children will temporary violate the integrity.
      renderObject.debugChildIntegrityEnabled = false;

      newChildren.keys.forEach(processElement); // !!

      if (_didUnderflow) {
        final int lastKey = _childElements.lastKey() ?? -1;
        final int rightBoundary = lastKey + 1;
        newChildren[rightBoundary] = _childElements[rightBoundary];
        processElement(rightBoundary);
      }
    } finally {
      _currentlyUpdatingChildIndex = null;
      renderObject.debugChildIntegrityEnabled = true;
    }
  }

  int _currentlyUpdatingChildIndex;

  /// Copied from [SliverMultiBoxAdaptorElement.debugAssertChildListLocked].
  @override
  bool debugAssertChildListLocked() {
    assert(_currentlyUpdatingChildIndex == null);
    return true;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.didAdoptChild].
  /// An assertion was removed to accept the dummy child element.
  @override
  void didAdoptChild(final RenderBox child) {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    // assert(_currentlyUpdatingChildIndex != null);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    final SliverMultiBoxAdaptorParentData childParentData =
        child.parentData as SliverMultiBoxAdaptorParentData;
    childParentData.index = _currentlyUpdatingChildIndex;
  }

  bool _didUnderflow = false;

  /// Copied from [SliverMultiBoxAdaptorElement.debugAssertChildListLocked].
  @override
  void setDidUnderflow(final bool value) {
    _didUnderflow = value;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.insertRenderObjectChild].
  /// This method has been changed in order to accept a null slot, in case of
  /// dummy child element.
  @override
  void insertRenderObjectChild(covariant RenderObject child, final int slot) {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if (slot == null) {
      renderObject.setupParentData(child);
      assert(renderObject.debugValidateChild(child));
      renderObject.adoptChild(child);
      return;
    }
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""

    assert(slot != null);
    assert(_currentlyUpdatingChildIndex == slot);
    assert(renderObject.debugValidateChild(child));
    renderObject.insert(child as RenderBox, after: _currentBeforeChild);
    assert(() {
      final SliverMultiBoxAdaptorParentData childParentData =
          child.parentData as SliverMultiBoxAdaptorParentData;
      assert(slot == childParentData.index);
      return true;
    }());
  }

  /// Copied from [SliverMultiBoxAdaptorElement.moveRenderObjectChild].
  /// This method has been changed to avoid calling the [AnimatedRenderSliverList] method
  /// when reorganizing rendered elements.
  @override
  void moveRenderObjectChild(
      covariant RenderObject child, final int oldSlot, final int newSlot) {
    assert(newSlot != null);
    assert(_currentlyUpdatingChildIndex == newSlot);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if (_remanagingChildren) return;
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    renderObject.move(child as RenderBox, after: _currentBeforeChild);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.removeRenderObjectChild].
  /// This method has been changed to remove the dummy child element correctly.
  @override
  void removeRenderObjectChild(covariant RenderObject child, final int slot) {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if ((child.parentData as SliverMultiBoxAdaptorParentData).index == null) {
      renderObject.dropChild(child);
      return;
    }
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""

    assert(_currentlyUpdatingChildIndex != null);
    renderObject.remove(child as RenderBox);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.visitChildren].
  @override
  void visitChildren(final ElementVisitor visitor) {
    // The toList() is to make a copy so that the underlying list can be modified by
    // the visitor:
    assert(!_childElements.values.any((Element child) => child == null));
    _childElements.values.cast<Element>().toList().forEach(visitor);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.debugVisitOnstageChildren].
  @override
  void debugVisitOnstageChildren(final ElementVisitor visitor) {
    _childElements.values.cast<Element>().where((Element child) {
      final SliverMultiBoxAdaptorParentData parentData =
          child.renderObject.parentData as SliverMultiBoxAdaptorParentData;
      double itemExtent;
      switch (renderObject.constraints.axis) {
        case Axis.horizontal:
          itemExtent = child.renderObject.paintBounds.width;
          break;
        case Axis.vertical:
          itemExtent = child.renderObject.paintBounds.height;
          break;
      }

      return parentData.layoutOffset != null &&
          parentData.layoutOffset <
              renderObject.constraints.scrollOffset +
                  renderObject.constraints.remainingPaintExtent &&
          parentData.layoutOffset + itemExtent >
              renderObject.constraints.scrollOffset;
    }).forEach(visitor);
  }

  /// Layouts the dummy child element.
  /// This method is called by [AnimatedRenderSliverList._measureIntervalSize]
  /// when calculating the size of a resizing interval.
  RenderBox _layoutDummyChild(
      final Widget widget, final BoxConstraints childConstraints) {
    owner.buildScope(this, () {
      _currentlyUpdatingChildIndex = null;
      _dummyChild = updateChild(_dummyChild, widget, null);
      _dummyChild.renderObject.layout(childConstraints, parentUsesSize: true);
    });
    return _dummyChild.renderObject as RenderBox;
  }

  /// Frees the dummy child element.
  /// This method is called by [AnimatedRenderSliverList._measureIntervalSize]
  /// when calculating the size of a resizing interval.
  void _dropDummyChild() {
    if (_dummyChild != null) {
      owner.buildScope(this, () {
        updateChild(_dummyChild, null, null);
        _dummyChild = null;
      });
    }
  }

  /// Copied from [SliverMultiBoxAdaptorElement.didStartLayout].
  @override
  void didStartLayout() {
    assert(debugAssertChildListLocked());
  }

  /// Copied from [SliverMultiBoxAdaptorElement.didFinishLayout].
  @override
  void didFinishLayout() {
    assert(debugAssertChildListLocked());
    final int firstIndex = _childElements.firstKey() ?? 0;
    final int lastIndex = _childElements.lastKey() ?? 0;
    widget.delegate.didFinishLayout(firstIndex, lastIndex);
  }
}

//---------------------------------------------------------------------------------------------
// _AnimationController
//---------------------------------------------------------------------------------------------

/// Helper class to control animations via [AnimationController].
class _AnimationController {
  AnimationController _controller;
  Animation<double> _animation;
  dynamic Function() _whenComplete;
  Curve _curve;

  _AnimationController({
    @required TickerProvider vsync,
    Duration duration,
    Curve curve,
    double startingControllerValue = 0.0,
    bool startNow = true,
    dynamic Function() whenComplete,
  }) {
    _curve = curve;
    _controller = AnimationController(
        vsync: vsync, duration: duration, value: startingControllerValue);
    if (curve != null)
      _animation = Tween(begin: 0.0, end: 1.0)
          .animate(CurvedAnimation(parent: _controller, curve: curve));
    _whenComplete = whenComplete;
    if (startNow) start();
  }

  _AnimationController.clone(
      TickerProvider vsync, _AnimationController controller)
      : this(
            vsync: vsync,
            duration: controller.duration,
            curve: controller._curve,
            startingControllerValue: controller.controllerValue,
            startNow: controller._controller.status == AnimationStatus.forward);

  void start() {
    if (_controller.status == AnimationStatus.dismissed) {
      assert(_animation != null);
      var future = _controller.forward();
      if (_whenComplete != null) future.whenComplete(_whenComplete);
    }
  }

  void restart(
      [Duration newDuration,
      Curve newCurve,
      dynamic Function() whenComplete,
      bool startNow = true]) {
    if (_controller.status != AnimationStatus.dismissed) _controller.reset();
    if (newDuration != null && newDuration != duration)
      _controller.duration = newDuration;
    if (newCurve != null && newCurve != _curve) {
      _curve = newCurve;
      _animation = Tween(begin: 0.0, end: 1.0)
          .animate(CurvedAnimation(parent: _controller, curve: newCurve));
    }
    _whenComplete = whenComplete;
    if (startNow) start();
  }

  get duration => _controller.duration;

  Animation<double> get animation => _animation;

  double get animationValue => _animation?.value ?? 0.0;

  double get controllerValue => _controller.value;

  void dispose() {
    _controller.dispose();
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedSliverChildBuilderDelegate
//---------------------------------------------------------------------------------------------

typedef AnimatedNullableIndexedWidgetBuilder = Widget Function(
    BuildContext context, int index, bool animating);

int _kDefaultSemanticIndexCallback(Widget _, int localIndex) => localIndex;

int _kChildCount() => null;

/// Class copied from [SliverChildBuilderDelegate].
///
/// The [builder] has been changed using a new callback that has a new animating parameter
/// indicating when to build the item in an animating state or not. For example, in animating
/// state you should avoid to use a key because a duplicate key exception will arise.
///
/// The [childCount] attribute has been replaced with a callback function. This callback
/// has to return the actual length of your list that will be changed after every change,
/// insertion, removal and replacement. A [null] value can be also returned, especially for
/// infinite lists.
class AnimatedSliverChildBuilderDelegate extends SliverChildDelegate {
  const AnimatedSliverChildBuilderDelegate(
    this.builder, {
    this.findChildIndexCallback,
    this.childCount = _kChildCount,
    this.addAutomaticKeepAlives = true,
    this.addRepaintBoundaries = true,
    this.addSemanticIndexes = true,
    this.semanticIndexCallback = _kDefaultSemanticIndexCallback,
    this.semanticIndexOffset = 0,
  })  : assert(builder != null),
        assert(addAutomaticKeepAlives != null),
        assert(addRepaintBoundaries != null),
        assert(addSemanticIndexes != null),
        assert(semanticIndexCallback != null);

  final AnimatedNullableIndexedWidgetBuilder builder;
  final int Function() childCount;
  final bool addAutomaticKeepAlives;
  final bool addRepaintBoundaries;
  final bool addSemanticIndexes;
  final int semanticIndexOffset;
  final SemanticIndexCallback semanticIndexCallback;
  final ChildIndexGetter findChildIndexCallback;

  /// Copied from [SliverChildBuilderDelegate.findIndexByIndex].
  @override
  int findIndexByKey(Key key) {
    if (findChildIndexCallback == null) return null;
    assert(key != null);
    Key childKey;
    if (key is _SaltedValueKey) {
      final _SaltedValueKey saltedValueKey = key;
      childKey = saltedValueKey.value;
    } else {
      childKey = key;
    }
    return findChildIndexCallback(childKey);
  }

  /// Copied from [SliverChildBuilderDelegate.build]. This method has been modified in order to
  /// handle the new animating attribute that indicates when an item should be build
  /// in an animating way or not. In addition, the new callback [AnimatedSliverChildBuilderDelegate.childCount]
  /// will be called instead of just an attribute.
  @override
  Widget build(BuildContext context, int index) {
    assert(builder != null);
    bool animating =
        (context as AnimatedSliverMultiBoxAdaptorElement)?._animating ?? true;
    int count = childCount?.call();
    if (index < 0 || (count != null && index >= count)) return null;
    Widget child;
    try {
      child = builder(context, index, animating);
    } catch (exception, stackTrace) {
      child = _createErrorWidget(exception, stackTrace);
    }
    if (child == null) {
      return null;
    }
    final Key key = child.key != null ? _SaltedValueKey(child.key) : null;
    if (addRepaintBoundaries) child = RepaintBoundary(child: child);
    if (addSemanticIndexes && !animating) {
      final int semanticIndex = semanticIndexCallback(child, index);
      if (semanticIndex != null)
        child = IndexedSemantics(
            index: semanticIndex + semanticIndexOffset, child: child);
    }
    if (addAutomaticKeepAlives && !animating)
      child = AutomaticKeepAlive(child: child);
    return KeyedSubtree(child: child, key: key);
  }

  /// Method not supported yet. Don't override it!
  @override
  double estimateMaxScrollOffset(
    int firstIndex,
    int lastIndex,
    double leadingScrollOffset,
    double trailingScrollOffset,
  ) {
    throw "'estimateMaxScrollOffset' method not supported yet.";
  }

  /// Copied from [SliverChildBuilderDelegate.estimatedChildCount]. The new callback
  /// [AnimatedSliverChildBuilderDelegate.childCount] will be called instead of just an attribute.
  @override
  int get estimatedChildCount => childCount();

  /// Copied from [SliverChildBuilderDelegate.shouldRebuild].
  @override
  bool shouldRebuild(
          covariant AnimatedSliverChildBuilderDelegate oldDelegate) =>
      true;
}

class _SaltedValueKey extends ValueKey<Key> {
  const _SaltedValueKey(Key key)
      : assert(key != null),
        super(key);
}

//---------------------------------------------------------------------------------------------
// Miscellaneous
//---------------------------------------------------------------------------------------------

/// Handle a list of all intervals [_Interval].
/// All intervals will be inserted oredered by its [_Interval.index].
class _IntervalList with IterableMixin {
  final List<_Interval> _list = List();

  @override
  Iterator get iterator => _list.iterator;

  _Interval operator [](final int index) => _list[index];

  int indexOf(final dynamic from) => _list.indexOfBinarySearch(from);

  void removeAt(final int index) => _list.removeAt(index);

  int insert(final _Interval interval) => _list.insertSorted(interval);

  _Interval intervalAt(final int index) => _list.binarySearch(index);

  /// Calculates the difference between the actual length of the underlying list and
  /// the actual length of the [AnimatedSliverList] that is animating.
  /// For example, consider a list of 10 elmenets; the user notifies an insertion
  /// of 5 elements somewhere. This sliver will start to animate with a resizing interval
  /// that occupies just one element. So, the sliver will count only 11 elements and
  /// this method will return -4.
  /// Only when the resizing interval becomes an inserting interval (or completes) this
  /// method will return 0 (as is no difference detected in the item count).
  int get itemCountAdjustment {
    int n = 0;
    for (final _Interval i in _list) {
      if (i.isInResizingState)
        n += 1 - i._insertCount;
      else if (i.isInRemovingState) n += i._removeCount - i._insertCount;
    }
    return n;
  }

  /// Start animations of those intervals that are still in the waiting state.
  /// If [coordinate] is [true], this method prioritizes the intervals
  /// in removal state first, then to those in resizing and changing state and finally
  /// to those in inserting state.
  void startAnimations(bool coordinate) {
    if (coordinate) {
      int state = 4;
      _list.forEach((i) {
        if (i.isInRemovingState)
          state = min(state, 1);
        else if (i.isInResizingState || i.isInChangingState)
          state = min(state, 2);
        else if (i.isInInsertingState) state = min(state, 3);
      });
      _list.forEach((i) {
        if (i.isInRemovingState && state == 1 ||
            (i.isInResizingState || i.isInChangingState) && state == 2 ||
            i.isInInsertingState && state == 3) i.startAnimation();
      });
    } else {
      _list.forEach((i) => i.startAnimation());
    }
  }

  void clear() {
    _list.forEach((i) => i.dispose());
    _list.clear();
  }

  @override
  String toString() => _list.toString();

  String toDebugString() {
    String s = "";
    for (_Interval i in _list) {
      if (s.isNotEmpty) s += "\n";
      s += i.toDebugString();
    }
    return s;
  }
}

enum _IntervalState {
  REMOVING,
  RESIZING,
  INSERTING,
  CHANGING,
  DISPOSED,
}

typedef _AnimatedListIntervalEventCallback = void Function(
    _Interval interval, bool beforeUpdatingState);

/// A single interval.
///
/// The interval changes its state during its life:
/// An interval arises with the [_IntervalState.REMOVING], [_IntervalState.RESIZING] (if no removal is needed) or
/// [_IntervalState.CANGING] state.
/// A removing interval always goes into the [_IntervalState.RESIZING] state.
/// A resizing interval can go into the [_IntervalState.INSERTING] state or directly into the [_IntervalState.DISPOSED]
/// (alias completed) state, if no insertion is needed.
/// A inserting interval, as well as a changing interval, always completes into the [_IntervalState.DISPOSED] state.
///
/// The interval takes into account two counts: [_removeCount] and [_insertCount]. The first indicates how many
/// items this interval will remove when it is in the removal animation, the latter indicates how many items
/// this intervals will insert when it is in the insert animation.
/// A changing interval has those counts always the same.
///
/// The [_index] is the actual position in the [AnimatedSliverList] (and not of the underlying list).
///
/// The values of [_fromSize], [_toSize] and [_lastSize] respectively indicate the initial size of the interval
/// when begin the resize animation, the final size at the end of the animation and the last rendered size.
/// The latter is used by [AnimatedSliverMultiBoxAdaptorElement._adjustScrollOffset] method to update the
/// scroll offset when a resize animation is in progress.
/// The [_currentSize] getter returns the current size based on the animation value.
/// [_fromSize] and [_toSize] can be initially [null]: that tells the renderer to measure them as this interval
/// goes into the [_IntervalState.RESIZING] state.
///
/// Animation won't start immediately. The [_waiting] attribute indicates that.
class _Interval extends Comparable {
  _IntervalState _state;

  int _index;

  int _insertCount, _removeCount;

  double _fromSize, _toSize, _lastSize;
  double get _currentSize {
    if (_fromSize != null && _toSize != null)
      return _fromSize + (_toSize - _fromSize) * animationValue;
    else
      return _fromSize != null && animationValue == 0.0 ? _fromSize : 0.0;
  }

  bool _waiting = true;

  double get animationValue =>
      _waiting ? 0.0 : (_controller?.animationValue ?? 0.0);

  _AnimationController _controller;

  /// This builder is used when the interval is in removing or changing state, and it can be
  /// used to build the item that is removing or changing, and therefore is no longer in the underlying list.
  NullableIndexedWidgetBuilder removeItemBuilder;

  /// Those callbacks are used to listen to all state changes.
  final _AnimatedListIntervalEventCallback onResizingCompleted,
      onRemovingCompleted,
      onInsertingCompleted,
      onChangingCompleted,
      onCompleted,
      onDisposed;

  bool get isInRemovingState => _state == _IntervalState.REMOVING;
  bool get isInResizingState => _state == _IntervalState.RESIZING;
  bool get isInInsertingState => _state == _IntervalState.INSERTING;
  bool get isInChangingState => _state == _IntervalState.CHANGING;

  bool get needsRemoveItemBuilder =>
      _state == _IntervalState.REMOVING || _state == _IntervalState.CHANGING;

  bool get needsInsertItemBuilder =>
      _state == _IntervalState.INSERTING || _state == _IntervalState.CHANGING;

  final AnimatedListAnimationSettings animationSettings;

  final TickerProvider vsync;

  /// This constructor creates a replacing interval. This interval will be created in the state [_IntervalState.REMOVING]
  /// or [_IntervalState.RESIZING] state (if no item should be removed).
  _Interval({
    @required this.vsync,
    @required this.animationSettings,
    @required int index,
    int insertCount = 0,
    int removeCount = 0,
    this.removeItemBuilder,
    this.onResizingCompleted,
    this.onRemovingCompleted,
    this.onInsertingCompleted,
    this.onCompleted,
    this.onDisposed,
  })  : _index = index,
        _removeCount = removeCount,
        _insertCount = insertCount,
        onChangingCompleted = null {
    assert(_removeCount != null && _insertCount != null);
    assert(_removeCount >= 0 && _insertCount >= 0);
    assert(_removeCount > 0 || _insertCount > 0);
    assert(removeItemBuilder != null || _removeCount == 0);

    if (_removeCount == 0) _fromSize = 0.0;
    if (_insertCount == 0) _toSize = 0.0;

    if (_removeCount > 0) {
      _changeState(_IntervalState.REMOVING);
    } else if (_insertCount > 0) {
      _changeState(_IntervalState.RESIZING);
    }
  }

  /// This constructor creates a changing interval. This interval will be created in the state [_IntervalState.CHANGING].
  _Interval.change({
    @required this.vsync,
    @required this.animationSettings,
    @required int index,
    @required int changeCount,
    @required this.removeItemBuilder,
    this.onChangingCompleted,
    this.onCompleted,
    this.onDisposed,
  })  : _index = index,
        onInsertingCompleted = null,
        onRemovingCompleted = null,
        onResizingCompleted = null {
    assert(changeCount != null);
    assert(changeCount > 0);
    assert(removeItemBuilder != null);

    _removeCount = changeCount;
    _insertCount = changeCount;

    _changeState(_IntervalState.CHANGING);
  }

  /// This constructor clones an existing interval.
  _Interval._clone(_Interval interval)
      : _state = interval._state,
        _lastSize = interval._lastSize,
        _index = interval._index,
        _insertCount = interval._insertCount,
        _removeCount = interval._removeCount,
        _fromSize = interval._fromSize,
        _toSize = interval._toSize,
        _waiting = interval._waiting,
        vsync = interval.vsync,
        removeItemBuilder = interval.removeItemBuilder,
        onResizingCompleted = interval.onResizingCompleted,
        onRemovingCompleted = interval.onRemovingCompleted,
        onInsertingCompleted = interval.onInsertingCompleted,
        onChangingCompleted = interval.onChangingCompleted,
        onCompleted = interval.onCompleted,
        onDisposed = interval.onDisposed,
        animationSettings = interval.animationSettings {
    assert(_state != _IntervalState.DISPOSED);
    if (interval._controller != null)
      _controller = _AnimationController.clone(vsync, interval._controller);
  }

  Animation<double> get animation => _waiting
      ? kAlwaysDismissedAnimation
      : (_controller?.animation ?? kAlwaysDismissedAnimation);

  void _changeState(_IntervalState newState) {
    assert(_state != _IntervalState.DISPOSED);
    _state = newState;
    _waiting = true;
  }

  void startAnimation() {
    if (!_waiting) return;
    Duration duration;
    Curve curve;
    void Function() whenComplete;
    switch (_state) {
      case _IntervalState.REMOVING:
        duration = animationSettings.removingDuration;
        curve = animationSettings.removingCurve;
        whenComplete = _onRemovingCompleted;
        break;
      case _IntervalState.RESIZING:
        duration = animationSettings.resizingDuration;
        curve = animationSettings.resizingCurve;
        whenComplete = _onResizingCompleted;
        break;
      case _IntervalState.INSERTING:
        duration = animationSettings.insertingDuration;
        curve = animationSettings.insertingCurve;
        whenComplete = _onInsertingCompleted;
        break;
      case _IntervalState.CHANGING:
        duration = animationSettings.changingDuration;
        curve = animationSettings.changingCurve;
        whenComplete = _onChangingCompleted;
        break;
      default:
        return;
    }
    if (_controller == null) {
      _controller ??= _AnimationController(
          vsync: vsync,
          duration: duration,
          curve: curve,
          whenComplete: whenComplete);
    } else {
      _controller.restart(duration, curve, whenComplete);
    }
    _waiting = false;
  }

  void _onRemovingCompleted() {
    if (_state == _IntervalState.DISPOSED) return;
    onRemovingCompleted?.call(this, true);
    _changeState(_IntervalState.RESIZING);
    onRemovingCompleted?.call(this, false);
  }

  void _onResizingCompleted() {
    if (_state == _IntervalState.DISPOSED) return;
    onResizingCompleted?.call(this, true);
    if (_insertCount > 0) {
      this.removeItemBuilder = null;
      _changeState(_IntervalState.INSERTING);
      onResizingCompleted?.call(this, false);
    } else {
      onResizingCompleted?.call(this, false);
      _onCompleted();
    }
  }

  void _onInsertingCompleted() {
    if (_state == _IntervalState.DISPOSED) return;
    onInsertingCompleted?.call(this, true);
    _onCompleted();
    onInsertingCompleted?.call(this, false);
  }

  void _onChangingCompleted() {
    if (_state == _IntervalState.DISPOSED) return;
    onChangingCompleted?.call(this, true);
    _onCompleted();
    onChangingCompleted?.call(this, false);
  }

  void _onCompleted() {
    if (_state == _IntervalState.DISPOSED) return;
    onCompleted?.call(this, true);
    dispose();
    onCompleted?.call(this, false);
  }

  /// Called by [AnimatedRenderSliverList._adjustInterval] method when this interval that was
  /// already in the resizing state changes its [_toSize].
  bool resize() {
    assert(isInResizingState);
    _fromSize = _currentSize;
    _toSize = (_insertCount == 0) ? 0.0 : null;
    if (_fromSize != null && _toSize != null && _fromSize.equals(_toSize)) {
      _onResizingCompleted();
      return true;
    } else {
      if (!_waiting) _controller?.restart(null, null, _onResizingCompleted);
    }
    return false;
  }

  /// Returns the count of items it is currently occupying in the [AnimatedSliverList].
  int get buildingItemCount {
    switch (_state) {
      case _IntervalState.REMOVING:
        return _removeCount;
      case _IntervalState.RESIZING:
        return 1;
      case _IntervalState.INSERTING:
      case _IntervalState.CHANGING:
        return _insertCount;
      default:
        return 0;
    }
  }

  /// Returns the difference between the count of items currently occupyied and the
  /// count of items it will occupy when this interval is completed.
  int get adjustingItemCount => _insertCount - buildingItemCount;

  @override
  String toString() {
    return "{index=$_index, removeCount:$_removeCount, insertCount=$_insertCount, state=$_state}";
  }

  String toDebugString() {
    String s;
    switch (_state) {
      case _IntervalState.REMOVING:
        s = "Rm";
        break;
      case _IntervalState.RESIZING:
        s = "Rz";
        break;
      case _IntervalState.INSERTING:
        s = "In";
        break;
      case _IntervalState.CHANGING:
        s = "Ch";
        break;
      case _IntervalState.DISPOSED:
        s = "Dd";
        break;
    }
    return "($s $_index R:$_removeCount I:$_insertCount)";
  }

  @override
  int compareTo(other) {
    if (other is int)
      return other - this._index;
    else if (other is _Interval) return other._index - this._index;
    throw "other in compareTo must be an integer or a ListInterval";
  }

  /// Eventually splits this interval.
  /// There are 4 possibilities:
  /// - if [leading] and [trailing] are both 0, this interval will be completed
  ///   and [null] is returned; the caller has to create a new interval for that
  ///   completly replaces it;
  /// - if only [leading] is 0, the interval will be narrowed down to [trailing] length
  ///   and [null] is returned; the [removeItemBuilder] will also be shifted
  ///   accordingly; the caller has to create a new interval that covers
  ///   the left side;
  /// - if only [trailing] is 0, the interval will be narrowed down to [leading] length
  ///   and [null] is returned; the caller has to create a new interval that covers
  ///   the right side;
  /// - if [leading] and [trailing] are both greater than 0, the interval will be narrowed
  ///   by the sum of [leading] and [trailing] and a new cloned [_Interval] will be
  ///   created and returned covering the right side; the caller has to create a new
  ///   interval that covers the middle part.
  _Interval split(TickerProvider vsync, int leading, int trailing) {
    assert(leading != null && trailing != null);
    assert(leading >= 0 && trailing >= 0);
    assert(leading + trailing <= this._insertCount);
    assert(this.isInChangingState || this.isInInsertingState);

    if (leading == 0 && trailing == 0) {
      _onCompleted();
      return null;
    }

    _Interval otherInterval;
    if (leading > 0 && trailing > 0) {
      otherInterval = _Interval._clone(this);
      if (otherInterval.isInChangingState) {
        otherInterval._removeCount = trailing;
      }
      otherInterval._insertCount = trailing;

      otherInterval._index += this._insertCount - trailing;

      if (otherInterval.needsRemoveItemBuilder) {
        otherInterval.removeItemBuilder = _shiftItemBuilder(
            otherInterval.removeItemBuilder, this._removeCount - trailing);
      }
    }

    if (leading > 0) {
      if (this.needsRemoveItemBuilder) {
        this._removeCount = leading;
      }
      this._insertCount = leading;
    } else {
      if (this.needsRemoveItemBuilder) {
        this.removeItemBuilder = _shiftItemBuilder(
            this.removeItemBuilder, this._removeCount - trailing);
      }

      this._index += this._insertCount - trailing;

      if (this.needsRemoveItemBuilder) {
        this._removeCount = trailing;
      }
      this._insertCount = trailing;
    }

    return otherInterval;
  }

  void dispose() {
    if (_state == _IntervalState.DISPOSED) return;
    onDisposed?.call(this, true);
    _state = _IntervalState.DISPOSED;
    if (_controller != null) {
      _controller.dispose();
      _controller = null;
    }
    onDisposed?.call(this, false);
  }
}

enum _BuildUpdateType {
  NEW_REMOVING_INTERVAL,
  NEW_RESIZING_INTERVAL,
  REMOVED_TO_RESIZING,
  RESIZED_TO_INSERTING,
  RESIZED_TO_DISPOSING,
  // INSERTED_TO_DISPOSING,
  // CHANGED_TO_DISPOSING,
}

class _BuildUpdate {
  final int index;
  final int removeCount;
  final int insertCount;
  final _BuildUpdateType type;

  _BuildUpdate({this.index, this.removeCount, this.insertCount, this.type});
}

// Copied from standard package.
class _WidgetTicker extends Ticker {
  _WidgetTicker(TickerCallback onTick, this._creator) : super(onTick);

  final AnimatedRenderSliverList _creator;

  @override
  void dispose() {
    _creator._removeTicker(this);
    super.dispose();
  }
}

// Return a Widget for the given Exception (copied from standard package).
Widget _createErrorWidget(Object exception, StackTrace stackTrace) {
  final FlutterErrorDetails details = FlutterErrorDetails(
    exception: exception,
    stack: stackTrace,
    library: 'widgets library',
    context: ErrorDescription('building'),
  );
  FlutterError.reportError(details);
  return ErrorWidget.builder(details);
}

NullableIndexedWidgetBuilder _shiftItemBuilder(
    final NullableIndexedWidgetBuilder builder, final int leading) {
  if (leading == 0 || builder == null) return builder;
  return (context, index) {
    return builder.call(context, index + leading);
  };
}

//

extension _DoubleExtension on double {
  bool equals(double d, [double precisionErrorTolerance = 1e-10]) {
    double c = this - d;
    return c < precisionErrorTolerance && c > -precisionErrorTolerance;
  }
}

extension _ListComparableExtension<T extends Comparable> on List<T> {
  int indexOfBinarySearch(dynamic element) {
    int min = 0;
    int max = this.length;
    while (min < max) {
      int mid = min + ((max - min) >> 1);
      int c = this[mid].compareTo(element);
      if (c < 0) {
        max = mid;
      } else if (c > 0) {
        min = mid + 1;
      } else {
        while (mid > 0) {
          if (this[mid - 1].compareTo(element) != 0) break;
          mid--;
        }
        return mid;
      }
    }
    return -1;
  }

  T binarySearch(dynamic element) {
    var i = indexOfBinarySearch(element);
    return (i >= 0) ? this[i] : null;
  }

  int nearestIndexOfBinarySearch(dynamic element) {
    int i = 0;
    int min = 0;
    int max = this.length;
    while (min < max) {
      final int mid = min + ((max - min) >> 1);
      int c = this[mid].compareTo(element);
      if (c < 0) {
        i = max = mid;
      } else if (c > 0) {
        i = min = mid + 1;
      } else {
        i = mid;
        break;
      }
    }
    return i;
  }

  int insertSorted(T element) {
    int i = nearestIndexOfBinarySearch(element);
    this.insert(i, element);
    return i;
  }
}
